from datetime import datetime, date
import hashlib
from google.appengine.api import users
from google.appengine.ext import db, search
import time
from process import settings

__author__ = 'IoS'

class SmartReferenceProperty(db.ReferenceProperty):
    def __init__(self,
                 reference_class=None,
                 verbose_name=None,
                 collection_name=None,
                 **attrs):
        super(SmartReferenceProperty, self).__init__(reference_class, verbose_name, collection_name, **attrs)

    def is_resolved(self, model_instance):
        if hasattr(model_instance, self._ReferenceProperty__id_attr_name()):
            reference_id = getattr(model_instance, self._ReferenceProperty__id_attr_name())
        else:
            reference_id = None
        if reference_id is not None:
            return getattr(model_instance, self._ReferenceProperty__resolved_attr_name()) is not None
        return False

class DictModel(object):
    SIMPLE_TYPES = (int, long, float, bool, dict, basestring, list)

    @classmethod
    def select_related(cls, entities = None):
        properties = cls.properties()
        references = cls.related_references if hasattr(cls, 'related_references') else [key for key, prop in properties.iteritems() if isinstance(prop, SmartReferenceProperty) or isinstance(prop, db.ReferenceProperty)]
        props = []
        for key in references:
            props.append(properties[key])
        return cls.prefetch_refprops(entities if entities is not None else cls.all(), *props)

    def to_dict(self):
        output = {}

        for key, prop in self.properties().iteritems():
            if key[0] == '_':
                continue

            value = None
            if isinstance(prop, SmartReferenceProperty):
                if prop.is_resolved(self):
                    value = getattr(self, key)
            else:
                value = getattr(self, key)

            if value is None or isinstance(value, self.SIMPLE_TYPES):
                output[key] = value
            elif isinstance(value, date) or isinstance(value, datetime):
                # Convert date/datetime to ms-since-epoch ("new Date()").
                ms = time.mktime(value.utctimetuple())
                ms += getattr(value, 'microseconds', 0) / 1000
                output[key] = int(ms)
            elif isinstance(value, db.GeoPt):
                output[key] = {'lat': value.lat, 'lon': value.lon}
            elif isinstance(value, DictModel):
                output[key] = value.to_dict()
            elif isinstance(value, db.Model):
                output[key] = value.key().id()
            elif isinstance(value, users.User):
                output[key] = {"id": value.user_id(), "name": value.nickname()}
            else:
                raise ValueError('cannot encode ' + repr(prop))

        output["id"] = self.key().id()

        return output

    @classmethod
    def prefetch_refprops(cls, entities, *props):
        fields = [(entity, prop) for entity in entities for prop in props]
        ref_keys = [prop.get_value_for_datastore(x) for x, prop in fields]
        ref_entities = dict((x.key(), x) for x in db.get(set(ref_keys)))
        for (entity, prop), ref_key in zip(fields, ref_keys):
            prop.__set__(entity, ref_entities[ref_key])
        return entities

class Unique(db.Model):
    @classmethod
    def check(cls, scope, value):
        def tx(scope, value):
            key_name = "U%s:%s" % (scope, value,)
            ue = Unique.get_by_key_name(key_name)
            if ue:
                raise UniqueConstraintViolation(scope, value)
            ue = Unique(key_name=key_name)
            ue.put()
        db.run_in_transaction(tx, scope, value)
    @classmethod
    def clear_key(cls, scope, value):
        def tx(scope, value):
            key_name = "U%s:%s" % (scope, value,)
            ue = Unique.get_by_key_name(key_name)
            if ue:
                ue.delete()
        db.run_in_transaction(tx, scope, value)
class UniqueConstraintViolation(Exception):
    def __init__(self, scope, value):
        super(UniqueConstraintViolation, self).__init__("Value '%s' is not unique within scope '%s'." % (value, scope, ))

######REFERENCES######
class Account(db.Model, DictModel):
    user = db.UserProperty(required=False, verbose_name="User")

class Action(db.Model, DictModel):
    USER_REGISTER = "User register"

    # project
    PROJECT_ADD = "Add project to read"
    PROJECT_CREATE = "Create new project"
    PROJECT_EDIT = "Edit project"
    PROJECT_DELETE = "Delete project"
    
    # add user to project
    USER_INVITE_TO_PROJECT = "Invite user into project as reader"
    USER_ACCEPT_TO_PROJECT = "User accepted invitation into project"
    USER_ROLE_IN_PROJECT = "Edit user's role in project"

    # file
    FILE_CREATE = "Create new file"
    FILE_PROPERTIES_EDIT = "Edit file properties"
    FILE_DELETE = "Delete file"
    FILE_EDIT_INSERT = "Edit file: insert"
    FILE_EDIT_DELETE = "Edit file: delete"

    # state
    #TODO

    ACTIONS = [USER_REGISTER,
                   PROJECT_ADD, PROJECT_CREATE, PROJECT_EDIT, PROJECT_DELETE,
                   USER_INVITE_TO_PROJECT, USER_ACCEPT_TO_PROJECT, USER_ROLE_IN_PROJECT,
                   FILE_CREATE, FILE_PROPERTIES_EDIT, FILE_DELETE, FILE_EDIT_INSERT, FILE_EDIT_DELETE]

    name = db.StringProperty(verbose_name="Action")

class ProjectType(search.SearchableModel, DictModel):
    name = db.StringProperty(required=True, verbose_name="Name")
    description = db.TextProperty(verbose_name="Brief description")
    _files = db.ListProperty(db.Key)

    @property
    def available_file_types(self):
        return FileType.get(self._files)

class FileType(search.SearchableModel, DictModel):
    name = db.StringProperty(required=True, verbose_name="Name")
    description = db.TextProperty(verbose_name="Brief description")
    extension = db.StringProperty(required=True, verbose_name="Extension")
    mimetype = db.StringProperty(required=False, default=None, verbose_name="Mime Type")

    @property
    def available_project_types(self):
        return ProjectType.gql("WHERE _files = :1", self.key())

######CURRENT STATE######
class Project(search.SearchableModel, db.Expando, DictModel):
    ACTIVE = 0
    SHELVED = 1
    ARCHIVED = 2
    STATUSES = [ACTIVE, SHELVED, ARCHIVED]

    name = db.StringProperty(required=True, verbose_name="Name")
    link_name = db.StringProperty(required=False, verbose_name="Link Name")
    description = db.TextProperty(verbose_name="Brief description")
    type = SmartReferenceProperty(ProjectType, required=True, verbose_name="Type")
    created = db.DateTimeProperty(auto_now_add=True, verbose_name="Date created")
    public = db.BooleanProperty(default=False, verbose_name="Public")
    owner = SmartReferenceProperty(Account, required=True, verbose_name="Owner")
    status = db.IntegerProperty(default=ACTIVE, choices=STATUSES, verbose_name="Status")

    related_references = ['type', 'owner']

    def format_link(self, n = None):
        return self.name.lower().replace(" ", "_") + ("_" + n if n is not None and len(n) else "")

    def put(self, **kwargs):
        link_name = None
        if self.link_name is None:
            n = ""
            ok = False
            while not ok:
                link_name = self.format_link(n)
                try:
                    Unique.check("Project:link_name", link_name)
                    ok = True
                except UniqueConstraintViolation:
                    pass
                n = (int(n if len(n) else 0) + 1).__str__()
            self.link_name = link_name
        try:
            return super(Project, self).put(**kwargs)
        except Exception:
            if link_name is not None:
                Unique.clear_key("Project:link_name", link_name)

class Participant(db.Model, DictModel):
    INVITED = 0
    READ = 1
    WRITE = 2
    CONFIGURE = 3
    ALLOWED = [INVITED, READ, WRITE, CONFIGURE]

    who = SmartReferenceProperty(Account, required=True, verbose_name="User")
    allowed = db.IntegerProperty(required=True, default=INVITED, choices=ALLOWED)
    project = SmartReferenceProperty(Project, required=True, verbose_name="Project")

    related_references = ['who', 'project']

class FileVersion(search.SearchableModel, db.Expando, DictModel):
    content = db.BlobProperty(required=False, verbose_name="Content")
    hash = db.StringProperty(required=False, verbose_name="Hash")
    size = db.IntegerProperty(required=True, verbose_name="Size")
    updated = db.DateTimeProperty(auto_now_add=True, verbose_name="Version date")
    owner = SmartReferenceProperty(Account, required=True, verbose_name="Owner")
    name = db.StringProperty(required=True, verbose_name="Name")
    category = db.StringProperty(default="", verbose_name="Category")
    type = SmartReferenceProperty(FileType, required=True, verbose_name="Type")

    related_references = ['type', 'owner']

    def compare(self, o):
        if not hasattr(o, "check_hash"):
            return False
        if self.check_hash() != o.check_hash():
            return False
        if not hasattr(o, "content"):
            return False
        return self.content == o.content

    def check_hash(self):
        if self.hash is None:
            self.hash = hashlib.md5(self.content).hexdigest()
        return self.hash

    def put(self, **kwargs):
        self.check_hash()
        if not self.size:
            self.size = len(self.content)
        return super(FileVersion, self).put(**kwargs)

class ProjectState(search.SearchableModel, db.Expando, DictModel):
    updated = db.DateTimeProperty(auto_now_add=True, verbose_name="Version date")
    version = db.IntegerProperty(required=True, default=0, verbose_name="Version number")
    previous_state = db.SelfReferenceProperty(verbose_name="Previous state")
    tag = db.StringProperty(default="master", verbose_name="Branch")
    comment = db.StringProperty(required=True, verbose_name="Comment")
    project = SmartReferenceProperty(Project, required=True, verbose_name="Project")

    related_references = ['project']

class ProjectStateMerge(db.Model, DictModel):
    project_state = SmartReferenceProperty(ProjectState, verbose_name="New state")
    source_state = SmartReferenceProperty(ProjectState, verbose_name="Source state", collection_name="sourcestatemerge_set")

    related_references = ['project_state', 'source_state']

class FileVersionState(db.Model, DictModel):
    project_state = SmartReferenceProperty(ProjectState, required=True, verbose_name="Commit")
    file_version = SmartReferenceProperty(FileVersion, required=True, verbose_name="File")

    related_references = ['project_state', 'file_version']

class ApprovedRenderer(db.Model, DictModel):
    project = SmartReferenceProperty(Project, verbose_name="Renderer project")
    when = db.DateTimeProperty(auto_now=True, verbose_name="Date approved")

    related_references = ['project']

class ApprovedEditor(db.Model, DictModel):
    project = SmartReferenceProperty(Project, verbose_name="Editor project")
    when = db.DateTimeProperty(auto_now=True, verbose_name="Date approved")

    related_references = ['project']

def fill():
    if not Account.all().count():
        internal = Account().put()
    else:
        internal = Account().all().filter('user =', None)[0]

    if not Action.all().count():
        for a in Action.ACTIONS:
            Action(name = a).put()

    pp_available_for = []
    pp2_available_for = []
    pp3_available_for = []
    pp4_available_for = []
    project_render = None
    file_js = None
    file_style = None
    file_png = None
    file_jpg = None
    neural_file = None
    h_file = None
    c_file = None
    cpp_file = None
    java_file = None
    project_editor = None
    project_flight = None
    # types
    if not FileType.all().count():
        # file types
        file_js = f = FileType(name = "JavaScript", description = "JS file", extension = "js", mimetype = 'application/javascript')
        f.put()
        pp2_available_for.append(f.key())
        file_style = f = FileType(name = "CSS", description = "CSS file", extension = "css", mimetype = 'text/css')
        f.put()
        pp2_available_for.append(f.key())
        file_png = f = FileType(name = "PNG", description = "PNG file", extension = "png", mimetype = 'image/png')
        f.put()
        pp3_available_for.append(f.key())
        file_jpg = f = FileType(name = "JPG", description = "JPG file", extension = "jpg", mimetype = 'image/jpg')
        f.put()
        pp3_available_for.append(f.key())
        neural_file = f = FileType(name = "Flight Process Markup Language", description = "Flight Process file", extension = "fpml", mimetype = 'text/xml')
        f.put()
        pp2_available_for.append(f.key())
        pp4_available_for.append(f.key())
        h_file = f = FileType(name = "H", description = "H file", extension = "h", mimetype = 'text/plain')
        f.put()
        pp2_available_for.append(f.key())
        c_file = f = FileType(name = "C", description = "C file", extension = "c", mimetype = 'text/plain')
        f.put()
        pp2_available_for.append(f.key())
        cpp_file = f = FileType(name = "C++", description = "C++ file", extension = "cpp", mimetype = 'text/plain')
        f.put()
        pp2_available_for.append(f.key())
        java_file = f = FileType(name = "Java", description = "Java file", extension = "java", mimetype = 'text/plain')
        f.put()
        pp2_available_for.append(f.key())

    if not ProjectType.all().count():
        # renderer
        project_render = p = ProjectType(name = "ProjectRender", description = "Project tree renderer. Connectable to current IDE. Must define project type to apply to and publication status, must define main source. Source file must contain function getRendererFactory(): Factory must implement method Renderer produce(desk, project, files), Renderer should implement method void filesChanged(newFileList).")
        p.put()
        pp_available_for.append(p.key())
        p._files.append(file_js.key())
        p._files.append(file_style.key())
        p._files.append(file_png.key())
        p._files.append(file_jpg.key())
        p.put()

        # editor
        project_editor = p = ProjectType(name = "Editor", description = "Editor. Connectable to current IDE. Must define file type to apply to and publication status, must define main source. Source file must contain function getEditorFactory(): Factory must implement method Editor produce(desk, project, file), Editor should implement method void fileChanged(newFile) and void activate().")
        p.put()
        pp_available_for.append(p.key())
        p._files.append(file_js.key())
        p.put()

        # neural network
        project_flight = p = ProjectType(name = "Flight Process", description = "Flight Process Project")
        p.put()
        pp_available_for.append(p.key())
        p._files.append(neural_file.key())
        p.put()

        # c/c++
        p = ProjectType(name = "C/C++", description = "C/C++ Project")
        p.put()
        pp_available_for.append(p.key())
        p._files.append(h_file.key())
        p._files.append(c_file.key())
        p._files.append(cpp_file.key())
        p.put()

        # java
        p = ProjectType(name = "Java", description = "Java Project")
        p.put()
        pp_available_for.append(p.key())
        p._files.append(java_file.key())
        p.put()

    # projects
    if not Project.all().count():
        pp = Project(name = "Project Renderer", description = "Basic tree renderer", type = project_render, public = True, owner = internal)
        pp.renderer_available_for = pp_available_for
        pp.main_renderer_source = "project-renderer.js"
        pp.put()
        ApprovedRenderer(project = pp).put()
        ps = ProjectState(version = 1, previous_state = None, comment = "Init", project = pp)
        ps.put()
        t = open(settings.ROOT_PATH + "/web/preload/project-renderer.js", "r")
        script = t.read()
        t.close()
        pfv = FileVersion(name = "project-renderer", type = file_js, content = script, size = len(script), owner = internal, version = 1)
        pfv.put()
        pfvs = FileVersionState(project_state = ps, file_version = pfv)
        pfvs.put()
        t = open(settings.ROOT_PATH + "/web/preload/project-renderer-styles.css", "r")
        script = t.read()
        t.close()
        pfv = FileVersion(name = "project-renderer", type = file_style, content = script, size = len(script), owner = internal, version = 1)
        pfv.put()
        pfvs = FileVersionState(project_state = ps, file_version = pfv)
        pfvs.put()
        t = open(settings.ROOT_PATH + "/web/preload/add-file-icon.png", "rb")
        script = t.read()
        t.close()
        pfv = FileVersion(name = "add-file-icon", type = file_png, content = script, size = len(script), owner = internal, version = 1)
        pfv.put()
        pfvs = FileVersionState(project_state = ps, file_version = pfv)
        pfvs.put()
        t = open(settings.ROOT_PATH + "/web/preload/delete-file-icon.png", "rb")
        script = t.read()
        t.close()
        pfv = FileVersion(name = "delete-file-icon", type = file_png, content = script, size = len(script), owner = internal, version = 1)
        pfv.put()
        pfvs = FileVersionState(project_state = ps, file_version = pfv)
        pfvs.put()

        pp2 = Project(name = "Text Editor", description = "Basic text editor", type = project_editor, public = True, owner = internal)
        pp2.editor_available_for = pp2_available_for
        pp2.main_editor_source = "text-editor.js"
        pp2.put()
        ApprovedEditor(project = pp2).put()
        ps2 = ProjectState(version = 1, previous_state = None, comment = "Init", project = pp2)
        ps2.put()
        t = open(settings.ROOT_PATH + "/web/preload/text-editor.js", "r")
        script = t.read()
        t.close()
        pfv2 = FileVersion(name = "text-editor", type = file_js, content = script, size = len(script), owner = internal, version = 1)
        pfv2.put()
        pfvs2 = FileVersionState(project_state = ps2, file_version = pfv2)
        pfvs2.put()

        pp3 = Project(name = "Image Editor", description = "Basic image editor", type = project_editor, public = True, owner = internal)
        pp3.editor_available_for = pp3_available_for
        pp3.main_editor_source = "image-editor.js"
        pp3.put()
        ApprovedEditor(project = pp3).put()
        ps3 = ProjectState(version = 1, previous_state = None, comment = "Init", project = pp3)
        ps3.put()
        t = open(settings.ROOT_PATH + "/web/preload/image-editor.js", "r")
        script = t.read()
        t.close()
        pfv3 = FileVersion(name = "image-editor", type = file_js, content = script, size = len(script), owner = internal, version = 1)
        pfv3.put()
        pfvs3 = FileVersionState(project_state = ps3, file_version = pfv3)
        pfvs3.put()

        pp4 = Project(name = "FlightProcess Editor", description = "Flight Process editor", type = project_editor, public = True, owner = internal)
        pp4.editor_available_for = pp4_available_for
        pp4.main_editor_source = "flightprocess-editor.js"
        pp4.put()
        ApprovedEditor(project = pp4).put()
        ps4 = ProjectState(version = 1, previous_state = None, comment = "Init", project = pp4)
        ps4.put()
        ##
        t = open(settings.ROOT_PATH + "/web/preload/flightprocess-editor.js", "r")
        script = t.read()
        t.close()
        pfv4 = FileVersion(name = "flightprocess-editor", type = file_js, content = script, size = len(script), owner = internal, version = 1)
        pfv4.put()
        pfvs4 = FileVersionState(project_state = ps4, file_version = pfv4)
        pfvs4.put()
        ##
        t = open(settings.ROOT_PATH + "/web/preload/add-node-icon.png", "rb")
        script = t.read()
        t.close()
        pfv4 = FileVersion(name = "add-node-icon", type = file_png, content = script, size = len(script), owner = internal, version = 1)
        pfv4.put()
        pfvs4 = FileVersionState(project_state = ps4, file_version = pfv4)
        pfvs4.put()
        ##
        t = open(settings.ROOT_PATH + "/web/preload/edit-edges-icon.png", "rb")
        script = t.read()
        t.close()
        pfv4 = FileVersion(name = "edit-edges-icon", type = file_png, content = script, size = len(script), owner = internal, version = 1)
        pfv4.put()
        pfvs4 = FileVersionState(project_state = ps4, file_version = pfv4)
        pfvs4.put()
        ##
        t = open(settings.ROOT_PATH + "/web/preload/edit-edge-icon.png", "rb")
        script = t.read()
        t.close()
        pfv4 = FileVersion(name = "edit-edge-icon", type = file_png, content = script, size = len(script), owner = internal, version = 1)
        pfv4.put()
        pfvs4 = FileVersionState(project_state = ps4, file_version = pfv4)
        pfvs4.put()
        ##
        t = open(settings.ROOT_PATH + "/web/preload/delete-node-icon.png", "rb")
        script = t.read()
        t.close()
        pfv4 = FileVersion(name = "delete-node-icon", type = file_png, content = script, size = len(script), owner = internal, version = 1)
        pfv4.put()
        pfvs4 = FileVersionState(project_state = ps4, file_version = pfv4)
        pfvs4.put()
        ##
        t = open(settings.ROOT_PATH + "/web/preload/delete-edge-icon.png", "rb")
        script = t.read()
        t.close()
        pfv4 = FileVersion(name = "delete-edge-icon", type = file_png, content = script, size = len(script), owner = internal, version = 1)
        pfv4.put()
        pfvs4 = FileVersionState(project_state = ps4, file_version = pfv4)
        pfvs4.put()
        ##
        t = open(settings.ROOT_PATH + "/web/preload/compile-icon.png", "rb")
        script = t.read()
        t.close()
        pfv4 = FileVersion(name = "compile-icon", type = file_png, content = script, size = len(script), owner = internal, version = 1)
        pfv4.put()
        pfvs4 = FileVersionState(project_state = ps4, file_version = pfv4)
        pfvs4.put()
        ##
        t = open(settings.ROOT_PATH + "/web/preload/flightprocess-editor-styles.css", "r")
        script = t.read()
        t.close()
        pfv4 = FileVersion(name = "flightprocess-editor-styles", type = file_style, content = script, size = len(script), owner = internal, version = 1)
        pfv4.put()
        pfvs4 = FileVersionState(project_state = ps4, file_version = pfv4)
        pfvs4.put()

fill()